@@ -2,18 +2,16 @@
  * \file Ellipsoid.hpp
  * \brief Header for GeographicLib::Ellipsoid class
  *
- * Copyright (c) Charles Karney (2012) <charles@karney.com> and licensed under
- * the MIT/X11 License.  For more information, see
- * http://geographiclib.sourceforge.net/
+ * Copyright (c) Charles Karney (2012-2023) <karney@alum.mit.edu> and licensed
+ * under the MIT/X11 License.  For more information, see
+ * https://geographiclib.sourceforge.io/
  **********************************************************************/
 
 #if !defined(GEOGRAPHICLIB_ELLIPSOID_HPP)
 #define GEOGRAPHICLIB_ELLIPSOID_HPP 1
 
 #include <GeographicLib/Constants.hpp>
-#include <GeographicLib/TransverseMercator.hpp>
-#include <GeographicLib/EllipticFunction.hpp>
-#include <GeographicLib/AlbersEqualArea.hpp>
+#include <GeographicLib/AuxLatitude.hpp>
 
 namespace GeographicLib {
 
@@ -21,16 +19,10 @@ namespace GeographicLib {
    * \brief Properties of an ellipsoid
    *
    * This class returns various properties of the ellipsoid and converts
-   * between various types of latitudes.  The latitude conversions are also
-   * possible using the various projections supported by %GeographicLib; but
-   * Ellipsoid provides more direct access (sometimes using private functions
-   * of the projection classes).  Ellipsoid::RectifyingLatitude,
-   * Ellipsoid::InverseRectifyingLatitude, and Ellipsoid::MeridianDistance
-   * provide functionality which can be provided by the Geodesic class.
-   * However Geodesic uses a series approximation (valid for abs \e f < 1/150),
-   * whereas Ellipsoid computes these quantities using EllipticFunction which
-   * provides accurate results even when \e f is large.  Use of this class
-   * should be limited to &minus;3 < \e f < 3/4 (i.e., 1/4 < b/a < 4).
+   * between various types of latitudes.  This is for the most part a thin
+   * wrapper on top of the AuxLatitude class which is called with \e exact =
+   * true so that the results are valid for arbitrary flattenings &minus;100 <
+   * \e f < 99/100 (i.e., 1/100 < b/a < 100).
    *
    * Example of use:
    * \include example-Ellipsoid.cpp
@@ -40,20 +32,10 @@ namespace GeographicLib {
   private:
     typedef Math::real real;
     static const int numit_ = 10;
-    static const real stol_;
-    real _a, _f, _f1, _f12, _e2, _e12, _n, _b;
-    TransverseMercator _tm;
-    EllipticFunction _ell;
-    AlbersEqualArea _au;
-    static real tand(real x) {
-      return
-        std::abs(x) == real(90) ? (x < 0 ?
-                                   - TransverseMercator::overflow_
-                                   : TransverseMercator::overflow_) :
-        std::tan(x * Math::degree<real>());
-    }
-    static real atand(real x)
-    { return std::atan(x) / Math::degree<real>(); }
+    real stol_;
+    real _a, _f,_b, _e2, _e12, _n;
+    AuxLatitude _aux;
+    real _rm, _c2;
 
   public:
     /** \name Constructor
@@ -61,13 +43,12 @@ namespace GeographicLib {
     ///@{
 
     /**
-     * Constructor for a ellipsoid with
+     * Constructor for an ellipsoid with
      *
      * @param[in] a equatorial radius (meters).
      * @param[in] f flattening of ellipsoid.  Setting \e f = 0 gives a sphere.
-     *   Negative \e f gives a prolate ellipsoid.  If \e f > 1, set flattening
-     *   to 1/\e f.
-     * @exception GeographicErr if \e a or (1 &minus; \e f ) \e a is not
+     *   Negative \e f gives a prolate ellipsoid.
+     * @exception GeographicErr if \e a or (1 &minus; \e f) \e a is not
      *   positive.
      **********************************************************************/
     Ellipsoid(real a, real f);
@@ -81,12 +62,12 @@ namespace GeographicLib {
      * @return \e a the equatorial radius of the ellipsoid (meters).  This is
      *   the value used in the constructor.
      **********************************************************************/
-    Math::real MajorRadius() const { return _a; }
+    Math::real EquatorialRadius() const { return _a; }
 
     /**
      * @return \e b the polar semi-axis (meters).
      **********************************************************************/
-    Math::real MinorRadius() const { return _b; }
+    Math::real PolarRadius() const { return _b; }
 
     /**
      * @return \e L the distance between the equator and a pole along a
@@ -108,7 +89,7 @@ namespace GeographicLib {
      *   a sphere with the same volume is cbrt(\e V / (4&pi;/3)).
      **********************************************************************/
     Math::real Volume() const
-    { return (4 * Math::pi<real>()) * Math::sq(_a) * _b / 3; }
+    { return (4 * Math::pi()) * Math::sq(_a) * _b / 3; }
     ///@}
 
     /** \name %Ellipsoid shape
@@ -171,7 +152,7 @@ namespace GeographicLib {
      * @param[in] phi the geographic latitude (degrees).
      * @return &beta; the parametric latitude (degrees).
      *
-     * The geographic latitude, &phi;, is the angle beween the equatorial
+     * The geographic latitude, &phi;, is the angle between the equatorial
      * plane and a vector normal to the surface of the ellipsoid.
      *
      * The parametric latitude (also called the reduced latitude), &beta;,
@@ -203,7 +184,7 @@ namespace GeographicLib {
      * @param[in] phi the geographic latitude (degrees).
      * @return &theta; the geocentric latitude (degrees).
      *
-     * The geocentric latitude, &theta;, is the angle beween the equatorial
+     * The geocentric latitude, &theta;, is the angle between the equatorial
      * plane and a line between the center of the ellipsoid and a point on the
      * ellipsoid.  For a sphere &theta; = &phi;.
      *
@@ -310,8 +291,10 @@ namespace GeographicLib {
      * defines the Mercator projection.  For a sphere &psi; =
      * sinh<sup>&minus;1</sup> tan &phi;.
      *
-     * &phi; must lie in the range [&minus;90&deg;, 90&deg;]; the
-     * result is undefined if this condition does not hold.
+     * &phi; must lie in the range [&minus;90&deg;, 90&deg;]; the result is
+     * undefined if this condition does not hold.  The value returned for &phi;
+     * = &plusmn;90&deg; is some (positive or negative) large but finite value,
+     * such that InverseIsometricLatitude returns the original value of &phi;.
      **********************************************************************/
     Math::real IsometricLatitude(real phi) const;
 
@@ -319,7 +302,8 @@ namespace GeographicLib {
      * @param[in] psi the isometric latitude (degrees).
      * @return &phi; the geographic latitude (degrees).
      *
-     * The returned value &phi; lies in [&minus;90&deg;, 90&deg;].
+     * The returned value &phi; lies in [&minus;90&deg;, 90&deg;].  For a
+     * sphere &phi; = tan<sup>&minus;1</sup> sinh &psi;.
      **********************************************************************/
     Math::real InverseIsometricLatitude(real psi) const;
     ///@}
@@ -396,9 +380,8 @@ namespace GeographicLib {
      *   section at latitude &phi; inclined at an angle \e azi to the
      *   meridian (meters).
      *
-     * &phi; must lie in the range [&minus;90&deg;, 90&deg;] and \e
-     * azi must lie in the range [&minus;540&deg;, 540&deg;); the
-     * result is undefined if either of conditions does not hold.
+     * &phi; must lie in the range [&minus;90&deg;, 90&deg;]; the result is
+     * undefined this condition does not hold.
      **********************************************************************/
     Math::real NormalCurvatureRadius(real phi, real azi) const;
     ///@}
@@ -460,7 +443,7 @@ namespace GeographicLib {
      * The returned value \e f lies in (&minus;&infin;, 1).
      **********************************************************************/
     static Math::real EccentricitySqToFlattening(real e2)
-    { return e2 / (std::sqrt(1 - e2) + 1); }
+    { using std::sqrt; return e2 / (sqrt(1 - e2) + 1); }
 
     /**
      * @param[in] f = (\e a &minus; \e b) / \e a, the flattening.
@@ -484,7 +467,7 @@ namespace GeographicLib {
      * The returned value \e f lies in (&minus;&infin;, 1).
      **********************************************************************/
     static Math::real SecondEccentricitySqToFlattening(real ep2)
-    { return ep2 / (std::sqrt(1 + ep2) + 1 + ep2); }
+    { using std::sqrt; return ep2 / (sqrt(1 + ep2) + 1 + ep2); }
 
     /**
      * @param[in] f = (\e a &minus; \e b) / \e a, the flattening.
@@ -507,8 +490,10 @@ namespace GeographicLib {
      * <i>e''</i> <sup>2</sup> should lie in (&minus;1, 1).
      * The returned value \e f lies in (&minus;&infin;, 1).
      **********************************************************************/
-    static Math::real ThirdEccentricitySqToFlattening(real epp2)
-    { return 2 * epp2 / (sqrt((1 - epp2) * (1 + epp2)) + 1 + epp2); }
+    static Math::real ThirdEccentricitySqToFlattening(real epp2) {
+      using std::sqrt;
+      return 2 * epp2 / (sqrt((1 - epp2) * (1 + epp2)) + 1 + epp2);
+    }
 
     /**
      * @param[in] f = (\e a &minus; \e b) / \e a, the flattening.
@@ -528,8 +513,7 @@ namespace GeographicLib {
      * A global instantiation of Ellipsoid with the parameters for the WGS84
      * ellipsoid.
      **********************************************************************/
-    static const Ellipsoid WGS84;
-
+    static const Ellipsoid& WGS84();
   };
 
 } // namespace GeographicLib
