@@ -2,9 +2,9 @@
  * \file MGRS.cpp
  * \brief Implementation for GeographicLib::MGRS class
  *
- * Copyright (c) Charles Karney (2008-2012) <charles@karney.com> and licensed
+ * Copyright (c) Charles Karney (2008-2022) <karney@alum.mit.edu> and licensed
  * under the MIT/X11 License.  For more information, see
- * http://geographiclib.sourceforge.net/
+ * https://geographiclib.sourceforge.io/
  **********************************************************************/
 
 #include <GeographicLib/MGRS.hpp>
@@ -14,39 +14,38 @@ namespace GeographicLib {
 
   using namespace std;
 
-  const Math::real MGRS::eps_ =
-    // 25 = ceil(log_2(2e7)) -- use half circumference here because northing
-    // 195e5 is a legal in the "southern" hemisphere.
-    pow(real(0.5), numeric_limits<real>::digits - 25);
-  const Math::real MGRS::angeps_ =
-    // 7 = ceil(log_2(90))
-    pow(real(0.5), numeric_limits<real>::digits - 7);
-  const string MGRS::hemispheres_ = "SN";
-  const string MGRS::utmcols_[3] = { "ABCDEFGH", "JKLMNPQR", "STUVWXYZ" };
-  const string MGRS::utmrow_ = "ABCDEFGHJKLMNPQRSTUV";
-  const string MGRS::upscols_[4] =
+  const char* const MGRS::hemispheres_ = "SN";
+  const char* const MGRS::utmcols_[] = { "ABCDEFGH", "JKLMNPQR", "STUVWXYZ" };
+  const char* const MGRS::utmrow_ = "ABCDEFGHJKLMNPQRSTUV";
+  const char* const MGRS::upscols_[] =
     { "JKLPQRSTUXYZ", "ABCFGHJKLPQR", "RSTUXYZ", "ABCFGHJ" };
-  const string MGRS::upsrows_[2] =
+  const char* const MGRS::upsrows_[] =
     { "ABCDEFGHJKLMNPQRSTUVWXYZ", "ABCDEFGHJKLMNP" };
-  const string MGRS::latband_ = "CDEFGHJKLMNPQRSTUVWX";
-  const string MGRS::upsband_ = "ABYZ";
-  const string MGRS::digits_ = "0123456789";
+  const char* const MGRS::latband_ = "CDEFGHJKLMNPQRSTUVWX";
+  const char* const MGRS::upsband_ = "ABYZ";
+  const char* const MGRS::digits_ = "0123456789";
+  const char* const MGRS::alpha_ = // Omit I+O
+    "ABCDEFGHJKLMNPQRSTUVWXYZabcdefghjklmnpqrstuvwxyz";
 
-  const int MGRS::mineasting_[4] =
+  const int MGRS::mineasting_[] =
     { minupsSind_, minupsNind_, minutmcol_, minutmcol_ };
-  const int MGRS::maxeasting_[4] =
+  const int MGRS::maxeasting_[] =
     { maxupsSind_, maxupsNind_, maxutmcol_, maxutmcol_ };
-  const int MGRS::minnorthing_[4] =
+  const int MGRS::minnorthing_[] =
     { minupsSind_, minupsNind_,
       minutmSrow_, minutmSrow_ - (maxutmSrow_ - minutmNrow_) };
-  const int MGRS::maxnorthing_[4] =
+  const int MGRS::maxnorthing_[] =
     { maxupsSind_, maxupsNind_,
       maxutmNrow_ + (maxutmSrow_ - minutmNrow_), maxutmNrow_ };
 
   void MGRS::Forward(int zone, bool northp, real x, real y, real lat,
                      int prec, std::string& mgrs) {
+    using std::isnan;           // Needed for Centos 7, ubuntu 14
+    // The smallest angle s.t., 90 - angeps() < 90 (approx 50e-12 arcsec)
+    // 7 = ceil(log_2(90))
+    static const real angeps = ldexp(real(1), -(Math::digits() - 7));
     if (zone == UTMUPS::INVALID ||
-        Math::isnan(x) || Math::isnan(y) || Math::isnan(lat)) {
+        isnan(x) || isnan(y) || isnan(lat)) {
       mgrs = "INVALID";
       return;
     }
@@ -71,16 +70,26 @@ namespace GeographicLib {
       // This isn't necessary...!  Keep y non-neg
       // if (!northp) y -= maxutmSrow_ * tile_;
     }
-    int
-      xh = int(floor(x)) / tile_,
-      yh = int(floor(y)) / tile_;
-    real
-      xf = x - tile_ * xh,
-      yf = y - tile_ * yh;
+    // The C++ standard mandates 64 bits for long long.  But
+    // check, to make sure.
+    static_assert(numeric_limits<long long>::digits >= 44,
+                  "long long not wide enough to store 10e12");
+    // Guard against floor(x * mult_) being computed incorrectly on some
+    // platforms.  The problem occurs when x * mult_ is held in extended
+    // precision and floor is inlined.  This causes tests GeoConvert1[678] to
+    // fail.  Problem reported and diagnosed by Thorkil Naur with g++ 10.2.0
+    // under Cygwin.
+    GEOGRAPHICLIB_VOLATILE real xx = x * mult_;
+    GEOGRAPHICLIB_VOLATILE real yy = y * mult_;
+    long long
+      ix = (long long)(floor(xx)),
+      iy = (long long)(floor(yy)),
+      m = (long long)(mult_) * (long long)(tile_);
+    int xh = int(ix / m), yh = int(iy / m);
     if (utmp) {
       int
         // Correct fuzziness in latitude near equator
-        iband = abs(lat) > angeps_ ? LatitudeBand(lat) : (northp ? 0 : -1),
+        iband = fabs(lat) < angeps ? (northp ? 0 : -1) : LatitudeBand(lat),
         icol = xh - minutmcol_,
         irow = UTMRow(iband, icol, yh % utmrowperiod_);
       if (irow != yh - (northp ? minutmNrow_ : maxutmSrow_))
@@ -95,32 +104,17 @@ namespace GeographicLib {
       int iband = (northp ? 2 : 0) + (eastp ? 1 : 0);
       mgrs1[z++] = upsband_[iband];
       mgrs1[z++] = upscols_[iband][xh - (eastp ? upseasting_ :
-                                         (northp ? minupsNind_ : minupsSind_))];
+                                         (northp ? minupsNind_ :
+                                          minupsSind_))];
       mgrs1[z++] = upsrows_[northp][yh - (northp ? minupsNind_ : minupsSind_)];
     }
     if (prec > 0) {
-      real mult = pow(real(base_), max(tilelevel_ - prec, 0));
-      int
-        ix = int(floor(xf / mult)),
-        iy = int(floor(yf / mult));
-      for (int c = min(prec, int(tilelevel_)); c--;) {
-        mgrs1[z + c] = digits_[ ix % base_ ];
-        ix /= base_;
-        mgrs1[z + c + prec] = digits_[ iy % base_ ];
-        iy /= base_;
-      }
-      if (prec > tilelevel_) {
-        xf -= floor(xf / mult);
-        yf -= floor(yf / mult);
-        mult = pow(real(base_), prec - tilelevel_);
-        ix = int(floor(xf * mult));
-        iy = int(floor(yf * mult));
-        for (int c = prec - tilelevel_; c--;) {
-          mgrs1[z + c + tilelevel_] = digits_[ ix % base_ ];
-          ix /= base_;
-          mgrs1[z + c + tilelevel_ + prec] = digits_[ iy % base_ ];
-          iy /= base_;
-        }
+      ix -= m * xh; iy -= m * yh;
+      long long d = (long long)(pow(real(base_), maxprec_ - prec));
+      ix /= d; iy /= d;
+      for (int c = prec; c--;) {
+        mgrs1[z + c       ] = digits_[ix % base_]; ix /= base_;
+        mgrs1[z + c + prec] = digits_[iy % base_]; iy /= base_;
       }
     }
     mgrs.resize(mlen);
@@ -140,8 +134,8 @@ namespace GeographicLib {
       // Here we do a more careful job using the band letter corresponding to
       // the actual latitude.
       ys /= tile_;
-      if (abs(ys) < 1)
-        lat = 0.9 * ys;         // accurate enough estimate near equator
+      if (fabs(ys) < 1)
+        lat = real(0.9) * ys;         // accurate enough estimate near equator
       else {
         real
           // The poleward bound is a fit from above of lat(x,y)
@@ -162,19 +156,19 @@ namespace GeographicLib {
     Forward(zone, northp, x, y, lat, prec, mgrs);
   }
 
-  void MGRS::Reverse(const std::string& mgrs,
+  void MGRS::Reverse(const string& mgrs,
                      int& zone, bool& northp, real& x, real& y,
                      int& prec, bool centerp) {
     int
       p = 0,
-      len = int(mgrs.size());
+      len = int(mgrs.length());
     if (len >= 3 &&
         toupper(mgrs[0]) == 'I' &&
         toupper(mgrs[1]) == 'N' &&
         toupper(mgrs[2]) == 'V') {
       zone = UTMUPS::INVALID;
       northp = false;
-      x = y = Math::NaN<real>();
+      x = y = Math::NaN();
       prec = -2;
       return;
     }
@@ -189,13 +183,13 @@ namespace GeographicLib {
     if (p > 0 && !(zone1 >= UTMUPS::MINUTMZONE && zone1 <= UTMUPS::MAXUTMZONE))
       throw GeographicErr("Zone " + Utility::str(zone1) + " not in [1,60]");
     if (p > 2)
-      throw GeographicErr("More than 2 digits_ at start of MGRS "
+      throw GeographicErr("More than 2 digits at start of MGRS "
                           + mgrs.substr(0, p));
     if (len - p < 1)
       throw GeographicErr("MGRS string too short " + mgrs);
     bool utmp = zone1 != UTMUPS::UPS;
     int zonem1 = zone1 - 1;
-    const string& band = utmp ? latband_ : upsband_;
+    const char* band = utmp ? latband_ : upsband_;
     int iband = Utility::lookup(band, mgrs[p++]);
     if (iband < 0)
       throw GeographicErr("Band letter " + Utility::str(mgrs[p-1]) + " not in "
@@ -203,7 +197,7 @@ namespace GeographicLib {
     bool northp1 = iband >= (utmp ? 10 : 2);
     if (p == len) {             // Grid zone only (ignore centerp)
       // Approx length of a degree of meridian arc in units of tile.
-      real deg = real(utmNshift_) / (90 * tile_);
+      real deg = real(utmNshift_) / (Math::qd * tile_);
       zone = zone1;
       northp = northp1;
       if (utmp) {
@@ -223,8 +217,8 @@ namespace GeographicLib {
       return;
     } else if (len - p < 2)
       throw GeographicErr("Missing row letter in " + mgrs);
-    const string& col = utmp ? utmcols_[zonem1 % 3] : upscols_[iband];
-    const string& row = utmp ? utmrow_ : upsrows_[northp1];
+    const char* col = utmp ? utmcols_[zonem1 % 3] : upscols_[iband];
+    const char* row = utmp ? utmrow_ : upsrows_[northp1];
     int icol = Utility::lookup(col, mgrs[p++]);
     if (icol < 0)
       throw GeographicErr("Column letter " + Utility::str(mgrs[p-1])
@@ -256,38 +250,36 @@ namespace GeographicLib {
     }
     int prec1 = (len - p)/2;
     real
-      unit = tile_,
-      x1 = unit * icol,
-      y1 = unit * irow;
+      unit = 1,
+      x1 = icol,
+      y1 = irow;
     for (int i = 0; i < prec1; ++i) {
-      unit /= base_;
+      unit *= base_;
       int
         ix = Utility::lookup(digits_, mgrs[p + i]),
         iy = Utility::lookup(digits_, mgrs[p + i + prec1]);
       if (ix < 0 || iy < 0)
         throw GeographicErr("Encountered a non-digit in " + mgrs.substr(p));
-      x1 += unit * ix;
-      y1 += unit * iy;
+      x1 = base_ * x1 + ix;
+      y1 = base_ * y1 + iy;
     }
     if ((len - p) % 2) {
       if (Utility::lookup(digits_, mgrs[len - 1]) < 0)
         throw GeographicErr("Encountered a non-digit in " + mgrs.substr(p));
       else
-        throw GeographicErr("Not an even number of digits_ in "
+        throw GeographicErr("Not an even number of digits in "
                             + mgrs.substr(p));
     }
     if (prec1 > maxprec_)
       throw GeographicErr("More than " + Utility::str(2*maxprec_)
-                          + " digits_ in "
-                          + mgrs.substr(p));
+                          + " digits in " + mgrs.substr(p));
     if (centerp) {
-      x1 += unit/2;
-      y1 += unit/2;
+      unit *= 2; x1 = 2 * x1 + 1; y1 = 2 * y1 + 1;
     }
     zone = zone1;
     northp = northp1;
-    x = x1;
-    y = y1;
+    x = (tile_ * x1) / unit;
+    y = (tile_ * y1) / unit;
     prec = prec1;
   }
 
@@ -295,15 +287,20 @@ namespace GeographicLib {
     // Limits are all multiples of 100km and are all closed on the lower end
     // and open on the upper end -- and this is reflected in the error
     // messages.  However if a coordinate lies on the excluded upper end (e.g.,
-    // after rounding), it is shifted down by eps_.  This also folds UTM
+    // after rounding), it is shifted down by eps.  This also folds UTM
     // northings to the correct N/S hemisphere.
+
+    // The smallest length s.t., 1.0e7 - eps() < 1.0e7 (approx 1.9 nm)
+    // 25 = ceil(log_2(2e7)) -- use half circumference here because
+    // northing 195e5 is a legal in the "southern" hemisphere.
+    static const real eps = ldexp(real(1), -(Math::digits() - 25));
     int
       ix = int(floor(x / tile_)),
       iy = int(floor(y / tile_)),
       ind = (utmp ? 2 : 0) + (northp ? 1 : 0);
     if (! (ix >= mineasting_[ind] && ix < maxeasting_[ind]) ) {
       if (ix == maxeasting_[ind] && x == maxeasting_[ind] * tile_)
-        x -= eps_;
+        x -= eps;
       else
         throw GeographicErr("Easting " + Utility::str(int(floor(x/1000)))
                             + "km not in MGRS/"
@@ -316,7 +313,7 @@ namespace GeographicLib {
     }
     if (! (iy >= minnorthing_[ind] && iy < maxnorthing_[ind]) ) {
       if (iy == maxnorthing_[ind] && y == maxnorthing_[ind] * tile_)
-        y -= eps_;
+        y -= eps;
       else
         throw GeographicErr("Northing " + Utility::str(int(floor(y/1000)))
                             + "km not in MGRS/"
@@ -336,7 +333,7 @@ namespace GeographicLib {
       } else if (!northp && iy >= maxutmSrow_) {
         if (y == maxutmSrow_ * tile_)
           // If on equator retain S hemisphere
-          y -= eps_;
+          y -= eps;
         else {
           northp = true;
           y -= utmNshift_;
@@ -346,15 +343,17 @@ namespace GeographicLib {
   }
 
   int MGRS::UTMRow(int iband, int icol, int irow) {
-    // Input is MGRS (periodic) row index and output is true row index.  Band
-    // index is in [-10, 10) (as returned by LatitudeBand).  Column index
-    // origin is easting = 100km.  Returns maxutmSrow_ if irow and iband are
-    // incompatible.  Row index origin is equator.
+    // Input is iband = band index in [-10, 10) (as returned by LatitudeBand),
+    // icol = column index in [0,8) with origin of easting = 100km, and irow =
+    // periodic row index in [0,20) with origin = equator.  Output is true row
+    // index in [-90, 95).  Returns maxutmSrow_ = 100, if irow and iband are
+    // incompatible.
 
     // Estimate center row number for latitude band
     // 90 deg = 100 tiles; 1 band = 8 deg = 100*8/90 tiles
-    real c = 100 * (8 * iband + 4)/real(90);
+    real c = 100 * (8 * iband + 4) / real(Math::qd);
     bool northp = iband >= 0;
+    // These are safe bounds on the rows
     //  iband minrow maxrow
     //   -10    -90    -81
     //    -9    -80    -72
@@ -382,10 +381,16 @@ namespace GeographicLib {
       maxrow = iband <   9 ?
       int(floor(c + real(4.4) - real(0.1) * northp)) :  94,
       baserow = (minrow + maxrow) / 2 - utmrowperiod_ / 2;
-    // Add maxutmSrow_ = 5 * utmrowperiod_ to ensure operand is positive
+    // Offset irow by the multiple of utmrowperiod_ which brings it as close as
+    // possible to the center of the latitude band, (minrow + maxrow) / 2.
+    // (Add maxutmSrow_ = 5 * utmrowperiod_ to ensure operand is positive.)
     irow = (irow - baserow + maxutmSrow_) % utmrowperiod_ + baserow;
-    if (irow < minrow || irow > maxrow) {
-      // Northing = 71*100km and 80*100km intersect band boundaries
+    if (!( irow >= minrow && irow <= maxrow )) {
+      // Outside the safe bounds, so need to check...
+      // Northing = 71e5 and 80e5 intersect band boundaries
+      //   y = 71e5 in scol = 2 (x = [3e5,4e5] and x = [6e5,7e5])
+      //   y = 80e5 in scol = 1 (x = [2e5,3e5] and x = [7e5,8e5])
+      // This holds for all the ellipsoids given in NGA.SIG.0012_2.0.0_UTMUPS.
       // The following deals with these special cases.
       int
         // Fold [-10,-1] -> [9,0]
@@ -394,6 +399,8 @@ namespace GeographicLib {
         srow = irow >= 0 ? irow : -irow - 1,
         // Fold [4,7] -> [3,0]
         scol = icol < 4 ? icol : -icol + 7;
+      // For example, the safe rows for band 8 are 71 - 79.  However row 70 is
+      // allowed if scol = [2,3] and row 80 is allowed if scol = [0,1].
       if ( ! ( (srow == 70 && sband == 8 && scol >= 2) ||
                (srow == 71 && sband == 7 && scol <= 2) ||
                (srow == 79 && sband == 9 && scol >= 1) ||
@@ -403,4 +410,101 @@ namespace GeographicLib {
     return irow;
   }
 
+  void MGRS::Decode(const string& mgrs,
+                    string& gridzone, string& block,
+                    string& easting, string& northing) {
+    string::size_type n = mgrs.length();
+    if (n >= 3 &&
+        toupper(mgrs[0]) == 'I' &&
+        toupper(mgrs[1]) == 'N' &&
+        toupper(mgrs[2]) == 'V') {
+      gridzone = mgrs.substr(0, 3);
+      block = easting = northing = "";
+      return;
+    }
+    string::size_type p0 = mgrs.find_first_not_of(digits_);
+    if (p0 == string::npos)
+      throw GeographicErr("MGRS::Decode: ref does not contain alpha chars");
+    if (!(p0 <= 2))
+      throw GeographicErr("MGRS::Decode: ref does not start with 0-2 digits");
+    string::size_type p1 = mgrs.find_first_of(alpha_, p0);
+    if (p1 != p0)
+      throw GeographicErr("MGRS::Decode: ref contains non alphanumeric chars");
+    p1 = min(mgrs.find_first_not_of(alpha_, p0), n);
+    if (!(p1 == p0 + 1 || p1 == p0 + 3))
+      throw GeographicErr("MGRS::Decode: ref must contain 1 or 3 alpha chars");
+    if (p1 == p0 + 1 && p1 < n)
+      throw GeographicErr("MGRS::Decode: ref contains junk after 1 alpha char");
+    if (p1 < n && (mgrs.find_first_of(digits_, p1) != p1 ||
+                   mgrs.find_first_not_of(digits_, p1) != string::npos))
+      throw GeographicErr("MGRS::Decode: ref contains junk at end");
+    if ((n - p1) & 1u)
+      throw GeographicErr("MGRS::Decode: ref must end with even no of digits");
+    // Here [0, p0) = initial digits; [p0, p1) = alpha; [p1, n) = end digits
+    gridzone = mgrs.substr(0, p0+1);
+    block = mgrs.substr(p0+1, p1 - (p0 + 1));
+    easting = mgrs.substr(p1, (n - p1) / 2);
+    northing = mgrs.substr(p1 + (n - p1) / 2);
+  }
+
+  void MGRS::Check() {
+    real lat, lon, x, y, t = tile_; int zone; bool northp;
+    UTMUPS::Reverse(31, true , 1*t,  0*t, lat, lon);
+    if (!( lon <   0 ))
+      throw GeographicErr("MGRS::Check: equator coverage failure");
+    UTMUPS::Reverse(31, true , 1*t, 95*t, lat, lon);
+    if (!( lat >  84 ))
+      throw GeographicErr("MGRS::Check: UTM doesn't reach latitude = 84");
+    UTMUPS::Reverse(31, false, 1*t, 10*t, lat, lon);
+    if (!( lat < -80 ))
+      throw GeographicErr("MGRS::Check: UTM doesn't reach latitude = -80");
+    UTMUPS::Forward(56,  3, zone, northp, x, y, 32);
+    if (!( x > 1*t ))
+      throw GeographicErr("MGRS::Check: Norway exception creates a gap");
+    UTMUPS::Forward(72, 21, zone, northp, x, y, 35);
+    if (!( x > 1*t ))
+      throw GeographicErr("MGRS::Check: Svalbard exception creates a gap");
+    UTMUPS::Reverse(0, true , 20*t, 13*t, lat, lon);
+    if (!( lat <  84 ))
+      throw
+        GeographicErr("MGRS::Check: North UPS doesn't reach latitude = 84");
+    UTMUPS::Reverse(0, false, 20*t,  8*t, lat, lon);
+    if (!( lat > -80 ))
+      throw
+        GeographicErr("MGRS::Check: South UPS doesn't reach latitude = -80");
+    // Entries are [band, x, y] either side of the band boundaries.  Units for
+    // x, y are t = 100km.
+    const short tab[] = {
+      0, 5,  0,   0, 9,  0,     // south edge of band 0
+      0, 5,  8,   0, 9,  8,     // north edge of band 0
+      1, 5,  9,   1, 9,  9,     // south edge of band 1
+      1, 5, 17,   1, 9, 17,     // north edge of band 1
+      2, 5, 18,   2, 9, 18,     // etc.
+      2, 5, 26,   2, 9, 26,
+      3, 5, 27,   3, 9, 27,
+      3, 5, 35,   3, 9, 35,
+      4, 5, 36,   4, 9, 36,
+      4, 5, 44,   4, 9, 44,
+      5, 5, 45,   5, 9, 45,
+      5, 5, 53,   5, 9, 53,
+      6, 5, 54,   6, 9, 54,
+      6, 5, 62,   6, 9, 62,
+      7, 5, 63,   7, 9, 63,
+      7, 5, 70,   7, 7, 70,   7, 7, 71,   7, 9, 71, // y = 71t crosses boundary
+      8, 5, 71,   8, 6, 71,   8, 6, 72,   8, 9, 72, // between bands 7 and 8.
+      8, 5, 79,   8, 8, 79,   8, 8, 80,   8, 9, 80, // y = 80t crosses boundary
+      9, 5, 80,   9, 7, 80,   9, 7, 81,   9, 9, 81, // between bands 8 and 9.
+      9, 5, 95,   9, 9, 95,     // north edge of band 9
+    };
+    const int bandchecks = sizeof(tab) / (3 * sizeof(short));
+    for (int i = 0; i < bandchecks; ++i) {
+      UTMUPS::Reverse(38, true, tab[3*i+1]*t, tab[3*i+2]*t, lat, lon);
+      if (!( LatitudeBand(lat) == tab[3*i+0] ))
+        throw GeographicErr("MGRS::Check: Band error, b = " +
+                            Utility::str(tab[3*i+0]) + ", x = " +
+                            Utility::str(tab[3*i+1]) + "00km, y = " +
+                            Utility::str(tab[3*i+2]) + "00km");
+    }
+  }
+
 } // namespace GeographicLib
