@@ -2,13 +2,15 @@
  * \file TransverseMercatorExact.cpp
  * \brief Implementation for GeographicLib::TransverseMercatorExact class
  *
- * Copyright (c) Charles Karney (2008-2011) <charles@karney.com> and licensed
+ * Copyright (c) Charles Karney (2008-2022) <karney@alum.mit.edu> and licensed
  * under the MIT/X11 License.  For more information, see
- * http://geographiclib.sourceforge.net/
+ * https://geographiclib.sourceforge.io/
  *
  * The relevant section of Lee's paper is part V, pp 67--101,
- * <a href="http://dx.doi.org/10.3138/X687-1574-4325-WM62">Conformal
- * Projections Based On Jacobian Elliptic Functions</a>.
+ * <a href="https://doi.org/10.3138/X687-1574-4325-WM62">Conformal
+ * Projections Based On Jacobian Elliptic Functions</a>;
+ * <a href="https://archive.org/details/conformalproject0000leel/page/92">
+ * borrow from archive.org</a>.
  *
  * The method entails using the Thompson Transverse Mercator as an
  * intermediate projection.  The projections from the intermediate
@@ -45,67 +47,36 @@ namespace GeographicLib {
 
   using namespace std;
 
-  const Math::real TransverseMercatorExact::tol_ =
-    numeric_limits<real>::epsilon();
-  const Math::real TransverseMercatorExact::tol1_ = real(0.1) * sqrt(tol_);
-  const Math::real TransverseMercatorExact::tol2_ = real(0.1) * tol_;
-  const Math::real TransverseMercatorExact::taytol_ = pow(tol_, real(0.6));
-  // Overflow value s.t. atan(overflow_) = pi/2
-  const Math::real TransverseMercatorExact::overflow_ = 1 / Math::sq(tol_);
-
   TransverseMercatorExact::TransverseMercatorExact(real a, real f, real k0,
                                                    bool extendp)
-    : _a(a)
-    , _f(f <= 1 ? f : 1/f)
+    : tol_(numeric_limits<real>::epsilon())
+    , tol2_(real(0.1) * tol_)
+    , taytol_(pow(tol_, real(0.6)))
+    , _a(a)
+    , _f(f)
     , _k0(k0)
     , _mu(_f * (2 - _f))        // e^2
     , _mv(1 - _mu)              // 1 - e^2
     , _e(sqrt(_mu))
     , _extendp(extendp)
-    , _Eu(_mu)
-    , _Ev(_mv)
+    , _eEu(_mu)
+    , _eEv(_mv)
   {
-    if (!(Math::isfinite(_a) && _a > 0))
-      throw GeographicErr("Major radius is not positive");
+    if (!(isfinite(_a) && _a > 0))
+      throw GeographicErr("Equatorial radius is not positive");
     if (!(_f > 0))
       throw GeographicErr("Flattening is not positive");
     if (!(_f < 1))
-      throw GeographicErr("Minor radius is not positive");
-    if (!(Math::isfinite(_k0) && _k0 > 0))
+      throw GeographicErr("Polar semi-axis is not positive");
+    if (!(isfinite(_k0) && _k0 > 0))
       throw GeographicErr("Scale is not positive");
   }
 
-  const TransverseMercatorExact
-  TransverseMercatorExact::UTM(Constants::WGS84_a<real>(),
-                               Constants::WGS84_f<real>(),
-                               Constants::UTM_k0<real>());
-
-  // tau = tan(phi), taup = sinh(psi)
-  Math::real TransverseMercatorExact::taup(real tau) const {
-    real
-      tau1 = Math::hypot(real(1), tau),
-      sig = sinh( _e * Math::atanh(_e * tau / tau1) );
-    return Math::hypot(real(1), sig) * tau - sig * tau1;
-  }
-
-  Math::real TransverseMercatorExact::taupinv(real taup) const {
-    real
-      // See comment in TransverseMercator.cpp about the initial guess
-      tau = taup/_mv,
-      stol = tol_ * max(real(1), abs(taup));
-    // min iterations = 1, max iterations = 2; mean = 1.94
-    for (int i = 0; i < numit_; ++i) {
-      real
-        tau1 = Math::hypot(real(1), tau),
-        sig = sinh( _e * Math::atanh(_e * tau / tau1 ) ),
-        taupa = Math::hypot(real(1), sig) * tau - sig * tau1,
-        dtau = (taup - taupa) * (1 + _mv * Math::sq(tau)) /
-        ( _mv * tau1 * Math::hypot(real(1), taupa) );
-      tau += dtau;
-      if (!(abs(dtau) >= stol))
-        break;
-    }
-    return tau;
+  const TransverseMercatorExact& TransverseMercatorExact::UTM() {
+    static const TransverseMercatorExact utm(Constants::WGS84_a(),
+                                             Constants::WGS84_f(),
+                                             Constants::UTM_k0());
+    return utm;
   }
 
   void TransverseMercatorExact::zeta(real /*u*/, real snu, real cnu, real dnu,
@@ -115,15 +86,18 @@ namespace GeographicLib {
     // atanh(snu * dnv) = asinh(snu * dnv / sqrt(cnu^2 + _mv * snu^2 * snv^2))
     // atanh(_e * snu / dnv) =
     //         asinh(_e * snu / sqrt(_mu * cnu^2 + _mv * cnv^2))
+    // Overflow value s.t. atan(overflow) = pi/2
+    static const real
+      overflow = 1 / Math::sq(numeric_limits<real>::epsilon());
     real
       d1 = sqrt(Math::sq(cnu) + _mv * Math::sq(snu * snv)),
       d2 = sqrt(_mu * Math::sq(cnu) + _mv * Math::sq(cnv)),
-      t1 = (d1 ? snu * dnv / d1 : (snu < 0 ? -overflow_ : overflow_)),
-      t2 = (d2 ? sinh( _e * Math::asinh(_e * snu / d2) ) :
-            (snu < 0 ? -overflow_ : overflow_));
+      t1 = (d1 != 0 ? snu * dnv / d1 : (signbit(snu) ? -overflow : overflow)),
+      t2 = (d2 != 0 ? sinh( _e * asinh(_e * snu / d2) ) :
+            (signbit(snu) ? -overflow : overflow));
     // psi = asinh(t1) - asinh(t2)
     // taup = sinh(psi)
-    taup = t1 * Math::hypot(real(1), t2) - t2 * Math::hypot(real(1), t1);
+    taup = t1 * hypot(real(1), t2) - t2 * hypot(real(1), t1);
     lam = (d1 != 0 && d2 != 0) ?
       atan2(dnu * snv, cnu * cnv) - _e * atan2(_e * cnu * snv, dnu * cnv) :
       0;
@@ -142,12 +116,12 @@ namespace GeographicLib {
   }
 
   // Starting point for zetainv
-  bool TransverseMercatorExact::zetainv0(real psi, real lam, real& u, real& v)
-    const {
+  bool TransverseMercatorExact::zetainv0(real psi, real lam,
+                                         real& u, real& v) const {
     bool retval = false;
-    if (psi < -_e * Math::pi<real>()/4 &&
-        lam > (1 - 2 * _e) * Math::pi<real>()/2 &&
-        psi < lam - (1 - _e) * Math::pi<real>()/2) {
+    if (psi < -_e * Math::pi()/4 &&
+        lam > (1 - 2 * _e) * Math::pi()/2 &&
+        psi < lam - (1 - _e) * Math::pi()/2) {
       // N.B. this branch is normally not taken because psi < 0 is converted
       // psi > 0 by Forward.
       //
@@ -159,14 +133,14 @@ namespace GeographicLib {
       // Inverting this gives:
       real
         psix = 1 - psi / _e,
-        lamx = (Math::pi<real>()/2 - lam) / _e;
-      u = Math::asinh(sin(lamx) / Math::hypot(cos(lamx), sinh(psix))) *
+        lamx = (Math::pi()/2 - lam) / _e;
+      u = asinh(sin(lamx) / hypot(cos(lamx), sinh(psix))) *
         (1 + _mu/2);
       v = atan2(cos(lamx), sinh(psix)) * (1 + _mu/2);
-      u = _Eu.K() - u;
-      v = _Ev.K() - v;
-    } else if (psi < _e * Math::pi<real>()/2 &&
-               lam > (1 - 2 * _e) * Math::pi<real>()/2) {
+      u = _eEu.K() - u;
+      v = _eEv.K() - v;
+    } else if (psi < _e * Math::pi()/2 &&
+               lam > (1 - 2 * _e) * Math::pi()/2) {
       // At w = w0 = i * Ev.K(), we have
       //
       //     zeta = zeta0 = i * (1 - _e) * pi/2
@@ -179,47 +153,51 @@ namespace GeographicLib {
       // When inverting this, we map arg(w - w0) = [-90, 0] to
       // arg(zeta - zeta0) = [-90, 180]
       real
-        dlam = lam - (1 - _e) * Math::pi<real>()/2,
-        rad = Math::hypot(psi, dlam),
+        dlam = lam - (1 - _e) * Math::pi()/2,
+        rad = hypot(psi, dlam),
         // atan2(dlam-psi, psi+dlam) + 45d gives arg(zeta - zeta0) in range
         // [-135, 225).  Subtracting 180 (since multiplier is negative) makes
         // range [-315, 45).  Multiplying by 1/3 (for cube root) gives range
         // [-105, 15).  In particular the range [-90, 180] in zeta space maps
         // to [-90, 0] in w space as required.
-        ang = atan2(dlam-psi, psi+dlam) - real(0.75) * Math::pi<real>();
+        ang = atan2(dlam-psi, psi+dlam) - real(0.75) * Math::pi();
       // Error using this guess is about 0.21 * (rad/e)^(5/3)
       retval = rad < _e * taytol_;
-      rad = Math::cbrt(3 / (_mv * _e) * rad);
+      rad = cbrt(3 / (_mv * _e) * rad);
       ang /= 3;
       u = rad * cos(ang);
-      v = rad * sin(ang) + _Ev.K();
+      v = rad * sin(ang) + _eEv.K();
     } else {
       // Use spherical TM, Lee 12.6 -- writing atanh(sin(lam) / cosh(psi)) =
       // asinh(sin(lam) / hypot(cos(lam), sinh(psi))).  This takes care of the
       // log singularity at zeta = Eu.K() (corresponding to the north pole)
-      v = Math::asinh(sin(lam) / Math::hypot(cos(lam), sinh(psi)));
+      v = asinh(sin(lam) / hypot(cos(lam), sinh(psi)));
       u = atan2(sinh(psi), cos(lam));
       // But scale to put 90,0 on the right place
-      u *= _Eu.K() / (Math::pi<real>()/2);
-      v *= _Eu.K() / (Math::pi<real>()/2);
+      u *= _eEu.K() / (Math::pi()/2);
+      v *= _eEu.K() / (Math::pi()/2);
     }
     return retval;
   }
 
   // Invert zeta using Newton's method
-  void TransverseMercatorExact::zetainv(real taup, real lam, real& u, real& v)
-    const  {
+  void TransverseMercatorExact::zetainv(real taup, real lam,
+                                        real& u, real& v) const  {
     real
-      psi = Math::asinh(taup),
-      scal = 1/Math::hypot(real(1), taup);
+      psi = asinh(taup),
+      scal = 1/hypot(real(1), taup);
     if (zetainv0(psi, lam, u, v))
       return;
-    real stol2 = tol2_ / Math::sq(max(psi, real(1)));
+    real stol2 = tol2_ / Math::sq(fmax(psi, real(1)));
     // min iterations = 2, max iterations = 6; mean = 4.0
-    for (int i = 0, trip = 0; i < numit_; ++i) {
+    for (int i = 0, trip = 0;
+         i < numit_ ||
+           GEOGRAPHICLIB_PANIC
+           ("Convergence failure in TransverseMercatorExact");
+         ++i) {
       real snu, cnu, dnu, snv, cnv, dnv;
-      _Eu.sncndn(u, snu, cnu, dnu);
-      _Ev.sncndn(v, snv, cnv, dnv);
+      _eEu.am(u, snu, cnu, dnu);
+      _eEv.am(v, snv, cnv, dnv);
       real tau1, lam1, du1, dv1;
       zeta(u, snu, cnu, dnu, v, snv, cnv, dnv, tau1, lam1);
       dwdzeta(u, snu, cnu, dnu, v, snv, cnv, dnv, du1, dv1);
@@ -245,8 +223,8 @@ namespace GeographicLib {
     // Lee 55.4 writing
     // dnu^2 + dnv^2 - 1 = _mu * cnu^2 + _mv * cnv^2
     real d = _mu * Math::sq(cnu) + _mv * Math::sq(cnv);
-    xi = _Eu.E(snu, cnu, dnu) - _mu * snu * cnu * dnu / d;
-    eta = v - _Ev.E(snv, cnv, dnv) + _mv * snv * cnv * dnv / d;
+    xi = _eEu.E(snu, cnu, dnu) - _mu * snu * cnu * dnu / d;
+    eta = v - _eEv.E(snv, cnv, dnv) + _mv * snv * cnv * dnv / d;
   }
 
   void TransverseMercatorExact::dwdsigma(real /*u*/,
@@ -265,23 +243,23 @@ namespace GeographicLib {
   }
 
   // Starting point for sigmainv
-  bool TransverseMercatorExact::sigmainv0(real xi, real eta, real& u, real& v)
-    const {
+  bool TransverseMercatorExact::sigmainv0(real xi, real eta,
+                                          real& u, real& v) const {
     bool retval = false;
-    if (eta > real(1.25) * _Ev.KE() ||
-        (xi < -real(0.25) * _Eu.E() && xi < eta - _Ev.KE())) {
+    if (eta > real(1.25) * _eEv.KE() ||
+        (xi < -real(0.25) * _eEu.E() && xi < eta - _eEv.KE())) {
       // sigma as a simple pole at w = w0 = Eu.K() + i * Ev.K() and sigma is
       // approximated by
       //
       // sigma = (Eu.E() + i * Ev.KE()) + 1/(w - w0)
       real
-        x = xi - _Eu.E(),
-        y = eta - _Ev.KE(),
+        x = xi - _eEu.E(),
+        y = eta - _eEv.KE(),
         r2 = Math::sq(x) + Math::sq(y);
-      u = _Eu.K() + x/r2;
-      v = _Ev.K() - y/r2;
-    } else if ((eta > real(0.75) * _Ev.KE() && xi < real(0.25) * _Eu.E())
-               || eta > _Ev.KE()) {
+      u = _eEu.K() + x/r2;
+      v = _eEv.K() - y/r2;
+    } else if ((eta > real(0.75) * _eEv.KE() && xi < real(0.25) * _eEu.E())
+               || eta > _eEv.KE()) {
       // At w = w0 = i * Ev.K(), we have
       //
       //     sigma = sigma0 = i * Ev.KE()
@@ -295,35 +273,39 @@ namespace GeographicLib {
       // arg(sigma - sigma0) = [-pi/2, pi/2]
       // mapping arg = [-pi/2, -pi/6] to [-pi/2, pi/2]
       real
-        deta = eta - _Ev.KE(),
-        rad = Math::hypot(xi, deta),
+        deta = eta - _eEv.KE(),
+        rad = hypot(xi, deta),
         // Map the range [-90, 180] in sigma space to [-90, 0] in w space.  See
         // discussion in zetainv0 on the cut for ang.
-        ang = atan2(deta-xi, xi+deta) - real(0.75) * Math::pi<real>();
+        ang = atan2(deta-xi, xi+deta) - real(0.75) * Math::pi();
       // Error using this guess is about 0.068 * rad^(5/3)
       retval = rad < 2 * taytol_;
-      rad = Math::cbrt(3 / _mv * rad);
+      rad = cbrt(3 / _mv * rad);
       ang /= 3;
       u = rad * cos(ang);
-      v = rad * sin(ang) + _Ev.K();
+      v = rad * sin(ang) + _eEv.K();
     } else {
       // Else use w = sigma * Eu.K/Eu.E (which is correct in the limit _e -> 0)
-      u = xi * _Eu.K()/_Eu.E();
-      v = eta * _Eu.K()/_Eu.E();
+      u = xi * _eEu.K()/_eEu.E();
+      v = eta * _eEu.K()/_eEu.E();
     }
     return retval;
   }
 
   // Invert sigma using Newton's method
-  void TransverseMercatorExact::sigmainv(real xi, real eta, real& u, real& v)
-    const {
+  void TransverseMercatorExact::sigmainv(real xi, real eta,
+                                         real& u, real& v) const {
     if (sigmainv0(xi, eta, u, v))
       return;
     // min iterations = 2, max iterations = 7; mean = 3.9
-    for (int i = 0, trip = 0; i < numit_; ++i) {
+    for (int i = 0, trip = 0;
+         i < numit_ ||
+           GEOGRAPHICLIB_PANIC
+           ("Convergence failure in TransverseMercatorExact");
+         ++i) {
       real snu, cnu, dnu, snv, cnv, dnv;
-      _Eu.sncndn(u, snu, cnu, dnu);
-      _Ev.sncndn(v, snv, cnv, dnv);
+      _eEu.am(u, snu, cnu, dnu);
+      _eEv.am(v, snv, cnv, dnv);
       real xi1, eta1, du1, dv1;
       sigma(u, snu, cnu, dnu, v, snv, cnv, dnv, xi1, eta1);
       dwdsigma(u, snu, cnu, dnu, v, snv, cnv, dnv, du1, dv1);
@@ -369,122 +351,116 @@ namespace GeographicLib {
   }
 
   void TransverseMercatorExact::Forward(real lon0, real lat, real lon,
-                                        real& x, real& y, real& gamma, real& k)
-    const {
-    lon = Math::AngDiff(Math::AngNormalize(lon0), Math::AngNormalize(lon));
+                                        real& x, real& y,
+                                        real& gamma, real& k) const {
+    lat = Math::LatFix(lat);
+    lon = Math::AngDiff(lon0, lon);
     // Explicitly enforce the parity
     int
-      latsign = (!_extendp && lat < 0) ? -1 : 1,
-      lonsign = (!_extendp && lon < 0) ? -1 : 1;
+      latsign = (!_extendp && signbit(lat)) ? -1 : 1,
+      lonsign = (!_extendp && signbit(lon)) ? -1 : 1;
     lon *= lonsign;
     lat *= latsign;
-    bool backside = !_extendp && lon > 90;
+    bool backside = !_extendp && lon > Math::qd;
     if (backside) {
       if (lat == 0)
         latsign = -1;
-      lon = 180 - lon;
+      lon = Math::hd - lon;
     }
     real
-      phi = lat * Math::degree<real>(),
-      lam = lon * Math::degree<real>(),
-      tau = tanx(phi);
+      lam = lon * Math::degree(),
+      tau = Math::tand(lat);
 
     // u,v = coordinates for the Thompson TM, Lee 54
     real u, v;
-    if (lat == 90) {
-      u = _Eu.K();
+    if (lat == Math::qd) {
+      u = _eEu.K();
       v = 0;
-    } else if (lat == 0 && lon == 90 * (1 - _e)) {
+    } else if (lat == 0 && lon == Math::qd * (1 - _e)) {
       u = 0;
-      v = _Ev.K();
+      v = _eEv.K();
     } else
-      zetainv(taup(tau), lam, u, v);
+      // tau = tan(phi), taup = sinh(psi)
+      zetainv(Math::taupf(tau, _e), lam, u, v);
 
     real snu, cnu, dnu, snv, cnv, dnv;
-    _Eu.sncndn(u, snu, cnu, dnu);
-    _Ev.sncndn(v, snv, cnv, dnv);
+    _eEu.am(u, snu, cnu, dnu);
+    _eEv.am(v, snv, cnv, dnv);
 
     real xi, eta;
     sigma(u, snu, cnu, dnu, v, snv, cnv, dnv, xi, eta);
     if (backside)
-      xi = 2 * _Eu.E() - xi;
+      xi = 2 * _eEu.E() - xi;
     y = xi * _a * _k0 * latsign;
     x = eta * _a * _k0 * lonsign;
 
-    if (lat == 90) {
+    if (lat == Math::qd) {
       gamma = lon;
       k = 1;
     } else {
       // Recompute (tau, lam) from (u, v) to improve accuracy of Scale
       zeta(u, snu, cnu, dnu, v, snv, cnv, dnv, tau, lam);
-      tau=taupinv(tau);
+      tau = Math::tauf(tau, _e);
       Scale(tau, lam, snu, cnu, dnu, snv, cnv, dnv, gamma, k);
-      gamma /= Math::degree<real>();
+      gamma /= Math::degree();
     }
     if (backside)
-      gamma = 180 - gamma;
+      gamma = Math::hd - gamma;
     gamma *= latsign * lonsign;
     k *= _k0;
   }
 
   void TransverseMercatorExact::Reverse(real lon0, real x, real y,
                                         real& lat, real& lon,
-                                        real& gamma, real& k)
-    const {
+                                        real& gamma, real& k) const {
     // This undoes the steps in Forward.
     real
       xi = y / (_a * _k0),
       eta = x / (_a * _k0);
     // Explicitly enforce the parity
     int
-      latsign = !_extendp && y < 0 ? -1 : 1,
-      lonsign = !_extendp && x < 0 ? -1 : 1;
-    xi *= latsign;
-    eta *= lonsign;
-    bool backside = !_extendp && xi > _Eu.E();
+      xisign = (!_extendp && signbit(xi)) ? -1 : 1,
+      etasign = (!_extendp && signbit(eta)) ? -1 : 1;
+    xi *= xisign;
+    eta *= etasign;
+    bool backside = !_extendp && xi > _eEu.E();
     if (backside)
-      xi = 2 * _Eu.E()- xi;
+      xi = 2 * _eEu.E()- xi;
 
     // u,v = coordinates for the Thompson TM, Lee 54
     real u, v;
-    if (xi == 0 && eta == _Ev.KE()) {
+    if (xi == 0 && eta == _eEv.KE()) {
       u = 0;
-      v = _Ev.K();
+      v = _eEv.K();
     } else
       sigmainv(xi, eta, u, v);
 
     real snu, cnu, dnu, snv, cnv, dnv;
-    _Eu.sncndn(u, snu, cnu, dnu);
-    _Ev.sncndn(v, snv, cnv, dnv);
+    _eEu.am(u, snu, cnu, dnu);
+    _eEv.am(v, snv, cnv, dnv);
     real phi, lam, tau;
-    if (v != 0 || u != _Eu.K()) {
+    if (v != 0 || u != _eEu.K()) {
       zeta(u, snu, cnu, dnu, v, snv, cnv, dnv, tau, lam);
-      tau = taupinv(tau);
+      tau = Math::tauf(tau, _e);
       phi = atan(tau);
-      lat = phi / Math::degree<real>();
-      lon = lam / Math::degree<real>();
+      lat = phi / Math::degree();
+      lon = lam / Math::degree();
       Scale(tau, lam, snu, cnu, dnu, snv, cnv, dnv, gamma, k);
-      gamma /= Math::degree<real>();
+      gamma /= Math::degree();
     } else {
-      tau = overflow_;
-      phi = Math::pi<real>()/2;
-      lat = 90;
+      lat = Math::qd;
       lon = lam = gamma = 0;
       k = 1;
     }
 
     if (backside)
-      lon = 180 - lon;
-    lon *= lonsign;
+      lon = Math::hd - lon;
+    lon *= etasign;
     lon = Math::AngNormalize(lon + Math::AngNormalize(lon0));
-    lat *= latsign;
+    lat *= xisign;
     if (backside)
-      y = 2 * _Eu.E() - y;
-    y *= _a * _k0 * latsign;
-    x *= _a * _k0 * lonsign;
-    if (backside)
-      gamma = 180 - gamma;
-    gamma *= latsign * lonsign;
+      gamma = Math::hd - gamma;
+    gamma *= xisign * etasign;
     k *= _k0;
   }
 
